/* circal 0.1 (c) 2004 Peter F. Stadler */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "alitypes.h"
#include "ali_io.h"
#include "scores.h"
#include "utils.h"

ALI *pair_align(ALI *A, ALI *B);
ALI *Rot_pair_align(ALI *A, ALI *B);
float  **Distances(int Nseq, ALI *R);
float  **RotDistances(int Nseq, ALI *R);
void free_ALI(ALI *C);
ALI *MultipleAlignment(int Nseq, ALI *R);
ALI *RotMultipleAlignment(int Nseq, ALI *R);
ALI *RotateAlignment(ALI *R, int r);
char *ReverseEntry(char *x);
ALI *circ_pair_align(ALI *A, ALI *B);

ALI *RotateAlignment(ALI *R, int r) {
  int i,j,k;
  ALI *X;
  X = (ALI *) space(sizeof(ALI)); 
  X->N = R->N;
  X->L = R->L;
  X->names = (char **) space(sizeof(char*)*X->N);
  X->ID =    (char **) space(sizeof(char*)*X->N);
  X->del =   (int *) space(sizeof(int)*(X->L+2));
  X->x   =   (int *) space(sizeof(int)*(X->L+2)); 
  X->type=   (char *) space(sizeof(char)*(X->L+2));
  for(i=0;i<R->N;i++) X->names[i] = strdup(R->names[i]);
  for(i=0;i<R->N;i++) X->ID[i]    = strdup(R->ID[i]);
  X->s = (char ***) space(sizeof(char**)*X->N);
  for(i=0;i<R->N;i++) X->s[i] = (char **) space(sizeof(char*)*(X->L+1)); 
  
  for(i=0;i<R->N;i++) {
    for(j=1;j<=R->L;j++) {
      if(r>=0) {
	k = j+r; if(k>R->L) k = k - R->L;
	/* printf(" + %2d %2d %2d\n",j,r,k); */
	X->s[i][k] = strdup(R->s[i][j]);
	if(i==0) {
	  X->del[k] = R->del[j];
	  X->x[k]   = R->x[j];
	  X->type[k]= R->type[j];
	}
      }
      else {
	k = X->L - j + r + 1;
	if(k<=0) k = k + R->L;
	if(k<=0) k = k + R->L;
	/* printf(" - %2d %2d %2d\n",j,r,k); */
	X->s[i][k] = ReverseEntry(R->s[i][j]);
	if(i==0) {
	  X->del[k] = R->del[j];
	  X->x[k]   = R->x[j];
	  X->type[k]= R->type[j];
	}
      }
    }
  }
  return X;
}

char *ReverseEntry(char *x) {
  int l;
  char *y;
  if(strcmp(x,"_____")==0) { y = strdup(x);   return y; } /*gap */
  if(x[0]=='-')            { y = strdup(x+1); return y; } /* reverse entry */
  /* regular entry */
  l = strlen(x);
  y = (char *) space(sizeof(char)*(l+2));
  y[0]='-';
  strcat(y,x);
  return y;
}

ALI *RotMultipleAlignment(int Nseq, ALI *R) {
  int i,j,k, im=0, jm=0;
  float **DD;
  float m, xi,xj;
  ALI **M, *PW;

  M = (ALI **) space(sizeof(ALI)*(Nseq));
  for(i=0;i<Nseq;i++) M[i] = R+i;  /* bloed, aber wahr */

  DD = RotDistances(Nseq, R);

 {
    FILE *fp;
    int p,q;
    fp=fopen("mat.dat","w");
    fprintf(fp,"%d\n",Nseq);
    for(p=0;p<Nseq;p++) {
      fprintf(fp,"%-30s ",R[p].names[0]);
      for(q=0;q<p;q++) fprintf(fp,"        ");
      fprintf(fp,"%7.1f ",0.0);
      for(q=p+1;q<Nseq;q++) 
	fprintf(fp,"%7.1f ",DD[p][q]);
      fprintf(fp,"\n");
    }
    fclose(fp);
  }

  for(k=0;k<Nseq-1;k++) {
    m=(float) INF;
    for(i=0;i<Nseq-1;i++) {
      if(M[i]!=NULL){ 
	for(j=i+1;j<Nseq;j++) {
	  if(M[j]!=NULL) { 
	    if(DD[i][j]<m) {
	      im = i; jm = j; m=DD[i][j];
	    }
	  }
	}
      }
    }
    fprintf(stderr,"%4d & %4d   //  %7.3f // %4d %4d \n",
	    im,jm,DD[im][jm],M[im]->L,M[jm]->L);
    xi = (float) M[im]->N / ( (float)(M[im]->N +M[jm]->N) );
    xj = (float) M[jm]->N / ( (float)(M[im]->N +M[jm]->N) );
    PW = Rot_pair_align(M[im],M[jm]);
    if(M[im]->N>1) free_ALI(M[im]);
    if(M[jm]->N>1) free_ALI(M[jm]);
    M[jm] = NULL;
       /* free intermediate multiple alignments */
    M[im] = PW;
    /* print_alignment(PW); 
       printf("\n--------------\n"); */
    for(i=0;i<im;i++) if(M[i]!=NULL)
      DD[i][im]    = xi*DD[i][im] + xj*DD[i][jm];
    for(j=im+1;j<Nseq;j++) if(M[j]!=NULL)
      DD[im][j] = xi*DD[im][j] + xj*DD[jm][j];
    for(j=0;j<im;j++) if(M[j]!=NULL)
      DD[im][j] = DD[j][im];
    for(i=im+1;i<Nseq;i++) if(M[i]!=NULL)
      DD[i][im] = DD[im][i];
  }

  PW = M[0];
  free(M);

  for(i=0;i<Nseq;i++) free(DD[i]);
  free(DD);
  
  return PW;
}

float  **RotDistances(int Nseq, ALI *R) {
  int i,j;
  float **D;
  D = (float **) space(Nseq*sizeof(float *));
  for(i=0;i<Nseq;i++) D[i] = (float *) space(Nseq*sizeof(float));

  for(i=0;i<Nseq-1;i++) {
    D[0][0] = 0.0;
    for(j=i+1;j<Nseq;j++) {
      ALI *Ctmp;
      Ctmp = Rot_pair_align(R+i, R+j);
      D[i][j] = (float) Ctmp->Score;
      D[j][i] = (float) Ctmp->Score;
      free_ALI(Ctmp);
      /* fprintf(stderr,"%6.2f ",D[i][j]); */
      fprintf(stderr,"*");
      fflush(stderr);
    }
    fprintf(stderr,"\n"); fflush(stderr);
  }
  return D;
}

ALI *Rot_pair_align(ALI *A, ALI *B) { 
  int r,s;
  int DD;
  ALI *Ar, *Bs,  *Copt;
  DD = INF;
  Copt = NULL;
  for(r=0;r<A->L;r++) { 
    Ar = RotateAlignment(A,r);
    for(s=- (B->L); s<B->L;s++) { 
      ALI *Ctmp;
      Bs = RotateAlignment(B,s);
      Ctmp = circ_pair_align(Ar, Bs);
      if(Ctmp->Score==INF) {
	free_ALI(Ctmp);
      }
      else if(Ctmp->Score < DD) {
	DD = Ctmp->Score;
	if(Copt!=NULL) free_ALI(Copt);
	Copt=Ctmp;
      }
      else{
	free_ALI(Ctmp);
      } 
      free_ALI(Bs);
    }
    free_ALI(Ar);
  }  
  return Copt;
}

void free_ALI(ALI *X) {
  int i,j;
  for(i=0;i<X->N; i++)  for(j=1;j<=X->L;j++) free(X->s[i][j]);
  for(i=0;i<X->N; i++) { 
    free(X->s[i]);
    free(X->names[i]);
    free(X->ID[i]);
  }
  free(X->del);
  free(X->x);
  free(X->type);
  free(X->s);
  free(X->names);
  free(X->ID);
  free(X);
}
  
/* under construction */

ALI *circ_pair_align(ALI *A, ALI *B) {
  int i,j,k,l,u;
  ALI *C;
  int **D;
  char *BT;
  int XXXX;
  
  C = (ALI*) space(sizeof(ALI));
  C->N = A->N + B->N;
  C->s      = (char ***) space(sizeof(char **)*(C->N));
  C->names  = (char **)  space(sizeof(char *)* (C->N));
  C->ID     = (char **)  space(sizeof(char *)* (C->N));

  for(i=0;i<A->N;i++) {
    C->names[i] = strdup(A->names[i]);
    C->ID[i]    = strdup(A->ID[i]);
  }
  for(i=0;i<B->N;i++) {
    C->names[A->N +i] = strdup(B->names[i]);
    C->ID[A->N +i]    = strdup(B->ID[i]);
  }

  XXXX = my_match(A,A->L,B,B->L);
  if(XXXX==INF) {  /* impossible alignment */
    C->L = 1;
    for(i=0;i<C->N;i++) { 
      C->s[i] = (char **) space(sizeof(char *)*2);
      C->s[i][1]= strdup("_____");
      /* C->del  = (int *) space(sizeof(int)*2);  */
      /* C->x    = (int *) space(sizeof(int)*2);  */     
      /* C->type = (char*) space(sizeof(char)*2); */ 
    }
    C->Score = INF;
    return C;
  }
  /* continue here only if the ends CAN match! */

  /*
  print_alignment(A);
  print_alignment(B);
  printf("&&&&&&&&&&&&&&&&&&\n");
  */
  
  D = (int **) space( sizeof(int*)*(A->L +1) );
  for(i=0;i<=A->L;i++) D[i] =  (int *) space( sizeof(int)*(B->L+1) );

  D[0][0]=0;
  for(i=1;i<=A->L;i++) D[i][0] = my_delete(A,1,i)*B->N;
  for(j=1;j<=B->L;j++) D[0][j] = my_delete(B,1,j)*A->N;

  for(i=1;i<=A->L;i++) {
    for(j=1;j<=B->L;j++) {
      int score;
      
      score = D[i-1][j-1] + my_match(A,i,B,j);

      for(k=1;k<=i;k++) {
	int d;
	d = D[k-1][j] + my_delete(A,k,i)*B->N;
	if(d<score) score =d;
      }

      for(k=1;k<=j;k++) {
	int d;
	d = D[i][k-1] + my_delete(B,k,j)*A->N;
	if(d<score) score =d;
      }
      D[i][j]=score;
    }
  }

  /* printf("score = %d\n",D[A->L][B->L]); */

  C->Score = D[A->L-1][B->L-1] + my_match(A,A->L,B,B->L);
  
  /* Backtracking */
  
  i=A->L-1; 
  j=B->L-1;
  u=0;
  BT = (char *) space(sizeof(char)*(A->L + B->L + 3));
  BT[u]='m';u++;  /* match at the end */
  while( (i>0)||(j>0) ) {
    int dd; int xxx; int l;
    xxx=0;
    if ((i>=1)&&(j>=1)) {
      /* printf("*M: %3d %d :: ",i,j); */
      dd = D[i-1][j-1]+my_match(A,i,B,j);
      if(dd==D[i][j]) {
	xxx=1;
	i--; j--;
	BT[u]='m';u++;
      }
    }
    if((i>=1)&&(xxx==0)) {
      /* printf("*I: %3d %d :: ",i,j); */
      for(k=i;k>0;k--) {
	dd = D[k-1][j] + my_delete(A,k,i)*B->N; 
	if(dd==D[i][j]) {
	  xxx=1;
	  for(l=k;l<=i;l++) { BT[u]='i';u++;  }
	  i= k-1;
	  k=-1;
	}
      }
    }
    if((j>=1)&&(xxx==0)) {
      /* printf("*D: %3d %d :: ",i,j); */
      for(k=j;k>0;k--) {
	dd = D[i][k-1] + my_delete(B,k,j)*A->N;
	if(dd==D[i][j]) {
	  xxx=1;
	  for(l=k;l<=j;l++) { BT[u]='d';u++;}
	  j= k-1;	  
	  k=-1;
	}
      }
    }
    /* printf("  %d -- %d %d\n",u, i,j ); */ 
    if(xxx==0) { printf("Do hots wos!\n"); exit(255); }
  }
  /* printf("\n"); */
  C->L = strlen(BT);
  C->del = (int *) space(sizeof(int)*(C->L+1));
  C->x   = (int *) space(sizeof(int)*(C->L+1));  
  C->type = (char *) space(sizeof(char)*(C->L+1));
  for(u=0;u<A->N+B->N;u++) C->s[u] = (char **) space(sizeof(char*)*(C->L+1));
  
  i=1;j=1;k=1;
  for(l=C->L-1;l>=0;l--) {
    switch(BT[l]) {
    case 'm' :
      for(u=0;u<A->N;u++) C->s[u][k]      = strdup(A->s[u][i]);
      for(u=0;u<B->N;u++) C->s[u+A->N][k] = strdup(B->s[u][j]);
      C->del[k] = A->del[i] + B->del[j];
      C->x[k]   = A->x[i]   + B->x[j];
      if((A->type[i])!=(B->type[j])) {
	int xx;
	fprintf(stderr,"unequal types - error!\n");
	fprintf(stderr,"A[%2d] %c // ",i,A->type[i]);
	for(xx=0;xx<A->N;xx++) fprintf(stderr,"%5s",A->s[xx][i]);
	fprintf(stderr,"\n");
	fprintf(stderr,"B[%2d] %c // ",j,B->type[j]);
	for(xx=0;xx<B->N;xx++) fprintf(stderr,"%5s",B->s[xx][j]);
	fprintf(stderr,"\n");
      }
	
      C->type[k] = A->type[i];
      i++; j++; k++;
      break;
    case 'd':
      for(u=0;u<A->N;u++) C->s[u][k]      = strdup("_____");
      for(u=0;u<B->N;u++) C->s[u+A->N][k] = strdup(B->s[u][j]);
      C->del[k]  = B->del[j];
      C->x[k]    = B->x[j]; 
      C->type[k] = B->type[j];
      j++;k++;
      break;
    case 'i':
      for(u=0;u<A->N;u++) C->s[u][k]      = strdup(A->s[u][i]);
      for(u=0;u<B->N;u++) C->s[u+A->N][k] = strdup("_____");
      C->del[k]  = A->del[i];
      C->x[k]    = A->x[i];
      C->type[k] = A->type[i];
      i++;k++;
      break;
    default:
      printf("This should never happen!\n");
    }
  }
  
  free(BT);
  for(i=0;i<=A->L;i++) free(D[i]);
  free(D);
  /* print_alignment(C); printf("%d\n",C->Score); */
			     
  return C;
}




