/* circal 0.1 (c) 2004 Peter F. Stadler */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "alitypes.h"
#include "ali_io.h"
#include "circal.h"
#include "scores.h"
#include "utils.h"

static void usage(char *name);

int main (int argc, char *argv[]) {
  int i,j,k,Nseq;
  ALI *R, *M;
  FILE *fp;
  char *scorefile;
  char *bname;
  char *infile, *nex_file, *ps_file;

  /* set defaults */
  scorefile = strdup("Mito.score");
  bname     = strdup("circal");
  infile    = NULL;
  
  for (i=1; i<argc; i++) {
    int r;
    char tstr[100];
    if (argv[i][0]=='-') 
      switch ( argv[i][1] ) {
      case 'S':  if (argv[i][2]!='\0') usage(argv[0]);
	if(i==argc-1) usage(argv[0]);
	r=sscanf(argv[++i], "%99s",tstr);
	free(scorefile);scorefile = strdup(tstr);
	if (!r) usage(argv[0]);
	break;
      case 'i':  if (argv[i][2]!='\0') usage(argv[0]);
	if(i==argc-1) usage(argv[0]);
	r=sscanf(argv[++i], "%s",tstr);
	infile = strdup(tstr);
	if (!r) usage(argv[0]);
	free(bname); bname = strdup(infile);
	k = strlen(bname);
	for(j=0;j<k;j++) if(bname[j]=='.') bname[j]='\0';
	break;
      default: usage(argv[0]);
      } 
  }

  ps_file =  (char *) space(sizeof(char)*(strlen(bname)+10));
  nex_file = (char *) space(sizeof(char)*(strlen(bname)+10));
  strcat(ps_file,bname);   strcat(ps_file,".ps");
  strcat(nex_file,bname);  strcat(nex_file,".nex");

  fp = stdin;
  if(infile!=NULL)  fp = fopen(infile,"r");
  R=read_Data(fp);
  if(infile!=NULL)  fclose(fp);

  Nseq=0; while(R[Nseq].N>=0) { Nseq++; }
  printf(">>> Read %d input records\n",Nseq);

  fp = fopen(scorefile,"r");
  if(fp==NULL) {
    fprintf(stderr,"Score file <%s> not found\n",scorefile);
    exit(255);
  }
  SSS = ReadScoreTypeFile(fp);
  fclose(fp);
  print_score_model(SSS);
  
  for(i=0;i<Nseq;i++) set_scores(R+i);
  for(i=0;i<Nseq;i++) print_alignment(R+i);
  
  M =  RotMultipleAlignment(Nseq,R);
  print_alignment(M);
  
  dump_NEX(M,nex_file);
  
  HAU_WECH(M,ps_file);
 
  free_ALI(M);
  for(k=0;k<Nseq;k++) {
    int i,j;
    ALI *X;
    X = R+k;
    for(i=0;i<X->N; i++)  for(j=1;j<=X->L;j++) free(X->s[i][j]);
    for(i=0;i<X->N; i++) { 
      free(X->s[i]);
      free(X->names[i]);
      free(X->ID[i]);
    }
    free(X->del);
    free(X->type);
    free(X->s);
    free(X->names);
    free(X->ID);
  }
  free(R);
  free(nex_file); free(ps_file); free(bname); free(scorefile);
  return 0;
}

static void usage(char *name) {
  fprintf(stderr,
	  "%s [-S scorefile] [-i inputfile]\n"
	  "%s [-S scorefile] < inputfile\n",name,name);
  exit(1);
}
