/* circal 0.1 (c) 2004 Peter F. Stadler */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "alitypes.h"
#include "utils.h"

ScoreTypes *SSS;

static int GOTAB[256];

int my_delete(ALI *A, int ia, int ib) {
  int score, go, k;
  score = 0; go=0;
  /* Determine 'open' score for the interval */
  for(k=ia;k<=ib;k++) {
    go = GOTAB[(int)A->type[k]] * A->x[k];
    if(go>score) score = go;  /* max score over all positions in putative gap*/
  }
  /* Add column contributions */
  for(k=ia;k<=ib;k++)  score += A->del[k];
  return score;
}

int my_match(ALI *A, int i, ALI *B, int j) {
  int score;
  int u,v, delA, delB;
  score = 0;

  for(u=0; u<A->N; u++) {
    for(v=0; v<B->N; v++) {
      if( (strcmp(A->s[u][i],"_____")!=0)&&
	  (strcmp(B->s[v][j],"_____")!=0)&&
	  (strcmp(A->s[u][i], B->s[v][j])!=0) ) {
	score = INF;
	return score;
      }
    }
  }
  delA=0;
  for(u=0; u<A->N; u++) 
    if(strcmp(A->s[u][i],"_____")==0) delA++;
  delB=0;
  for(v=0; v<B->N; v++)
    if(strcmp(B->s[v][j],"_____")==0) delB++;
  score = ( A->N - delA )*delB + ( B->N - delB )*delA;

  return score;
}


ScoreTypes *ReadScoreTypeFile(FILE *fp) {
  int i,x;
  char *line;
  ScoreTypes *SS;
  SS = (ScoreTypes *) space(sizeof(ScoreTypes));
  line = get_line(fp);
  while(line[0]=='#') { free(line); line=get_line(fp); }
  /*   (void) sscanf(line,"%d",&x); SS->Gap_Open=x;
       free(line); line=get_line(fp);
  */
  while(line[0]=='#') { free(line); line=get_line(fp); } 
  (void) sscanf(line,"%d",&x); SS->n = x;
  SS->a  = (char **) space(sizeof(char*)*SS->n);
  SS->s  = (int   *) space(sizeof(int)*SS->n);
  SS->go = (int   *) space(sizeof(int)*SS->n);  
  SS->type = (char*) space(sizeof(char)*SS->n);
  free(line);
  for(i=0;i<SS->n;i++) {
    int j,l;
    line = get_line(fp);
    while(line[0]=='#') { free(line); line=get_line(fp); } 
    (void) sscanf(line,"%c %d %d",&(SS->type[i]),&(SS->go[i]),&(SS->s[i]));
    l = strlen(line);
    if(i==0) SS->a[i] = NULL;
    if(i>0) {
      j=0; 
      while(line[j]!=':') { line[j]=' '; j++; }
      SS->a[i] = strdup(line+j+1);
    }
    free(line);
  }
  for(i=0;i<SS->n;i++) GOTAB[ (int) SS->type[i] ] = SS->go[i];
  
  return SS;
}

void set_scores(ALI *R){
  int i,k,score;
  char tp;
  char token[100];
  
  for(k=1;k<=R->L;k++) {
    token[0]=' ';
    token[1]='\0';
    if(R->s[0][k][0]=='-') {
      strcat(token,R->s[0][k]+1); /* printf("%s\n",token); */
    }
    else strcat(token,R->s[0][k]);
    strcat(token," ");
    score = SSS->s[0];
    tp = 'x';
    for(i=1;i<SSS->n;i++) {
      if(strstr(SSS->a[i],token)!=NULL) {
	score = SSS->s[i];
	tp    = SSS->type[i];
	/* printf("^%s^ in '%s' [%c]\n",token, SSS->a[i], tp); */
	i=SSS->n;
      }
    }
    R->del[k] = score;
    R->x[k]   = 1;
    R->type[k]= tp;
  }
}

void print_score_model(ScoreTypes *s) {
  int i;
  printf(">>> Scoring scheme:\n");
  printf("Type   open  extend  members\n");
  for(i=1;i<s->n;i++) {
    printf("[%d]  %c %5d %5d : %s\n", i, s->type[i],s->go[i], s->s[i],
	   s->a[i]);
  }
  printf("[%d]  %c %5d %5d : (default)\n",0,s->type[0], s->go[0], s->s[0]);
  printf("\n");
}
